/*
 * engine.c - URLfilterDB
 *
 * ufdbGuard is copyrighted (C) 2005-2016 by URLfilterDB with all rights reserved.
 *
 * Parts of ufdbGuard are based on squidGuard.
 *
 * RCS $Id: engine.c,v 1.19 2016/12/11 23:38:15 root Exp root $
 */


#include "ufdb.h"
#include "sg.h"
#include "ufdblib.h"
#include "ufdbdb.h"
#include "ufdbchkport.h"

#include <pthread.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define UFDB_DEBUG_ACL_ACCESS 0


UFDB_GCC_ALIGN64 pthread_rwlock_t  TheDynamicSourcesLock;
UFDB_GCC_ALIGN64 pthread_rwlock_t  TheDynamicCategoriesLock;
UFDB_GCC_ALIGN64 int __dummy0;


void UFDBaclEngineInit( void )
{
   pthread_rwlockattr_t mylock_attr;

   pthread_rwlockattr_init( &mylock_attr );

#if HAVE_PTHREAD_RWLOCK_PREFER_WRITER_NP
   pthread_rwlockattr_setkind_np( &mylock_attr, PTHREAD_RWLOCK_PREFER_WRITER_NP );
#else
   ufdbLogMessage( "WARNING: could not set writer lock preference for the dynamic source lock.  *****\n"
                   "This may delay updates for sources with execuserlist and execdomainlist   *****" );
#endif

   pthread_rwlock_init( &TheDynamicSourcesLock, &mylock_attr );
   pthread_rwlock_init( &TheDynamicCategoriesLock, &mylock_attr );
}


/*
 *  called by main loop: find a source definition based on the user ident/domain/ip
 */
UFDB_GCC_HOT
struct Source * UFDBfindSource( 
   struct Source *    allSrcs, 
   char *             net, 
   struct SquidInfo * si )
{
   struct Source *    s;
   struct Ip *        ip;
   int                foundip;
   int                founduser;
   int                founddomain;
   unsigned int       i, octet, op;
 
   if (UFDBglobalFatalError  ||  UFDBglobalReconfig)
   {
#if UFDB_DEBUG_ACL_ACCESS
      ufdbLogMessage( "   UFDBfindSource: fatal-error or reconfig: returning NULL" );
#endif
      return NULL;
   }

   if (allSrcs == NULL)
   {
      /* allSrcs may be NULL if there is a configuration with only the default rule */
      return NULL;
   }

   si->matchedBy[0] = '\0';
   octet = 0;
   if (net != NULL)
   {
      if (sgConvDot(net,&op) != NULL)
         octet = op;
   }

#if UFDB_DEBUG_ACL_ACCESS
   if (UFDBglobalDebug > 2)
      ufdbLogMessage( "UFDBfindSource   id '%s'  net '%s'  0x%08x",
                      si->ident, 
		      net == NULL ? "nonet" : net,
		      octet );
#endif

   for (s = allSrcs;  s != NULL;  s = s->next)
   {
      if (s->active == 0)
         continue;

      foundip = founduser = founddomain = 0;

#if UFDB_DEBUG_ACL_ACCESS
      if (UFDBglobalDebug > 2)
	 ufdbLogMessage( "   UFDBfindSource: trying source %s evaluate-%s", s->name, s->evaluationMethod == UFDB_EVAL_OR ? "or" : "and" );
#endif

      if (s->ip != NULL)
      {
         if (net == NULL)
	    ;
         else 
         {
	    for (ip = s->ip;  ip != NULL;  ip = ip->next)
	    {
	       if (!ip->net_is_set)
	          continue;

	       if (ip->type == SG_IPTYPE_RANGE) 
	       {
	          if (octet >= ip->net  &&  octet <= ip->mask)
	          {
	             strcpy( si->matchedBy, "ip-range" );
		     if (s->evaluationMethod == UFDB_EVAL_OR)
		     {
#if UFDB_DEBUG_ACL_ACCESS
			if (UFDBglobalDebug > 2)
			   ufdbLogMessage( "   UFDBfindSource: evaluate-or: ip \"%s\" found in list of ip ranges of source %s", 
					   net, s->name );
#endif
			return s;
		     }
		     else
		     {
#if UFDB_DEBUG_ACL_ACCESS
			if (UFDBglobalDebug > 2)
			   ufdbLogMessage( "   UFDBfindSource: evaluate-and: ip \"%s\" found in list of ip ranges of source %s", 
					   net, s->name );
#endif
		        foundip = 1;
			break;
		     }
	          }
	       }
	       else				/* CIDR or HOST */
	       { 
	          i = octet & ip->mask;
	          if (i == ip->net)
	          {
		     strcpy( si->matchedBy, "ip-subnet" );
		     if (s->evaluationMethod == UFDB_EVAL_OR)
		     {
#if UFDB_DEBUG_ACL_ACCESS
			if (UFDBglobalDebug > 2)
			   ufdbLogMessage( "   UFDBfindSource: evaluate-or: ip \"%s\" found in list of ip subnets of source %s", 
					   net, s->name );
#endif
			return s;
		     }
		     else
		     {
#if UFDB_DEBUG_ACL_ACCESS
			if (UFDBglobalDebug > 2)
			   ufdbLogMessage( "   UFDBfindSource: evaluate-and: ip \"%s\" found in list of ip subnets of source %s", 
					   net, s->name );
#endif
		        foundip = 1;
			break;
		     }
	          }
	       }
	    }
         }
      }
      else
      {
#if UFDB_DEBUG_ACL_ACCESS
	 if (UFDBglobalDebug > 2)
	    ufdbLogMessage( "   UFDBfindSource: source %s has no ip list", s->name );
#endif
         foundip = 1;
      }

#if UFDB_DEBUG_ACL_ACCESS
      if (UFDBglobalDebug > 2)
	 ufdbLogMessage( "   UFDBfindSource  '%s'  userDb = %08lx  foundip %d", 
	                 s->name==NULL ? "nosrc" : s->name,  (unsigned long) s->userDb, foundip  );
#endif

      if (s->userDb != NULL)
      {
	 if (si->ident[0] == '\0'  ||  (si->ident[0] == '-'  &&  si->ident[1] == '\0'))
	 {
	    founduser = 0;
	 }
	 else 
	 {
	    struct UserQuotaInfo * q = NULL;
	    int ret;

#if UFDB_DEBUG_ACL_ACCESS
	    if (UFDBglobalDebug > 2)
	       ufdbLogMessage( "   UFDBfindSource: matching user \"%s\" to userlist of \"%s\"", 
	                       si->ident, s->name==NULL ? "NULL" : s->name );
#endif
	    if (s->userDb->type == SGDBTYPE_EXECUSERLIST)
	    {
	       ret = pthread_rwlock_rdlock( &TheDynamicSourcesLock );
	       if (ret != 0)
		  ufdbLogError( "UFDBfindSource: pthread_rwlock_rdlock TheDynamicSourcesLock failed with code %d", ret );
	    }

	    if (UFDBmemDBfind( (struct UFDBmemDB *) s->userDb->dbcp, si->ident, (char **) (char *) &q ) == 1)
	    {
	       strcpy( si->matchedBy, "username" );
	       founduser = 1;
	    }

	    if (s->userDb->type == SGDBTYPE_EXECUSERLIST)
	    {
	       ret = pthread_rwlock_unlock( &TheDynamicSourcesLock );
	       if (ret != 0)
		  ufdbLogError( "UFDBfindSource: pthread_rwlock_unlock TheDynamicSourcesLock failed with code %d", ret );
	    }

	    if (founduser)
	    {
	       if (s->evaluationMethod == UFDB_EVAL_OR)
	       {
#if UFDB_DEBUG_ACL_ACCESS
		  if (UFDBglobalDebug > 2)
		     ufdbLogMessage( "   UFDBfindSource: evaluate-or: user \"%s\" found in userlist of source \"%s\"", 
				     si->ident, s->name==NULL ? "NULL" : s->name );
#endif
		  return s;
	       }
	       else
	       {
#if UFDB_DEBUG_ACL_ACCESS
		  if (UFDBglobalDebug > 2)
		     ufdbLogMessage( "   UFDBfindSource: evaluate-and: user \"%s\" found in userlist of source \"%s\"", 
				     si->ident, s->name==NULL ? "NULL" : s->name );
#endif
	       }
	    }
	 }
      }
      else
      {
#if UFDB_DEBUG_ACL_ACCESS
	 if (UFDBglobalDebug > 2)
	    ufdbLogMessage( "   UFDBfindSource: source %s has no user list", s->name );
#endif
         founduser = 1;
      }

      if (s->domainDb != NULL)
      {
         if (si->srcDomain[0] == '\0'  ||  (si->srcDomain[0] == '-'  &&  si->srcDomain[1] == '\0'))
	    founddomain = 0;
         else 
         {
#if UFDB_DEBUG_ACL_ACCESS
	    if (UFDBglobalDebug > 2)
	       ufdbLogMessage( "   UFDBfindSource: matching domain \"%s\" to list of domains of source %s", 
	                       si->srcDomain, s->name );
#endif
	    if (UFDBmemDBfindDomain( (struct UFDBmemDB *) s->domainDb, si->srcDomain ))
	    {
	       strcpy( si->matchedBy, "hostname" );
	       founddomain = 1;
	       if (founddomain)
	       {
	          if (s->evaluationMethod == UFDB_EVAL_OR)
		  {
#if UFDB_DEBUG_ACL_ACCESS
		     if (UFDBglobalDebug > 2)
			ufdbLogMessage( "   UFDBfindSource: evaluate-or: domain \"%s\" found in list of domains of source %s", 
					si->srcDomain, s->name );
#endif
		     return s;
		  }
		  else
		  {
#if UFDB_DEBUG_ACL_ACCESS
		     if (UFDBglobalDebug > 2)
			ufdbLogMessage( "   UFDBfindSource: evaluate-and: domain \"%s\" found in list of domains of source %s", 
					si->srcDomain, s->name );
#endif
		  }
	       }
            }
         }
      }
      else
      {
#if UFDB_DEBUG_ACL_ACCESS
	 if (UFDBglobalDebug > 2)
	    ufdbLogMessage( "   UFDBfindSource: source %s has no domain list", s->name );
#endif
         founddomain = 1;
      }

      if (founduser && foundip && founddomain)		/* evaluate-and */
      {
#if UFDB_DEBUG_ACL_ACCESS
	 if (UFDBglobalDebug > 2)
	    ufdbLogMessage( "   UFDBfindSource: evaluate-and: source \"%s\" matched", 
		            s->name==NULL ? "NULL" : s->name );
#endif
	 return s;
      }
   }

   return NULL;
}


/*
 * called by main loop: find the first active ACL that matches a source
 */
UFDB_GCC_HOT
struct Acl * UFDBfindACLbySource( 
   struct Source *     source,
   struct SquidInfo *  si  )
{
   struct Acl *    acl;

   if (source == NULL)
   {
#if UFDB_DEBUG_ACL_ACCESS
      if (UFDBglobalDebug)
#else
      if (UFDBglobalDebug > 1)
#endif
	 ufdbLogMessage( "   UFDBfindACLbySource: source=NULL returning defaultACL" );

      /* the username and IP address are not used to make the decision to use the ACL "default"
       * but for logging and reporting purposes, the username or IP address is given here.
       */
      if (si->ident[0] != '-'  &&  si->ident[1] != '\0')
      {
         strcpy( si->matchedBy, "username" );
      }
      else if (si->src[0] != '-'  &&  si->src[1] != '\0')
      {
         strcpy( si->matchedBy, "ip" );
      }
      else if (si->srcDomain[0] != '-'  &&  si->srcDomain[1] != '\0')
      {
         strcpy( si->matchedBy, "hostname" );
      }
      else
	 strcpy( si->matchedBy, "default" );

      return defaultAcl;
   }

   for (acl = AclList;  acl != NULL;  acl = acl->next)
   {
      if (acl->source == source  &&  acl->active)
      {
#if UFDB_DEBUG_ACL_ACCESS
	 if (UFDBglobalDebug)
	    ufdbLogMessage( "   UFDBfindACLbySource: active ACL for source '%s' found", source->name );
#endif

	 return acl;
      }
   }

#if UFDB_DEBUG_ACL_ACCESS
   if (UFDBglobalDebug)
      ufdbLogMessage( "   UFDBfindACLbySource: no active ACL for source '%s' found, returning defaultAcl", source->name );
#endif

   return defaultAcl;
}


void UFDBmakeRedirectString( 
   struct SquidInfo * si,
   struct Acl * acl,
   struct Category * category,
   char * redirect )
{
   char * format;

   if (category != NULL  &&  category->redirect != NULL)
      format = category->redirect;
   else if (acl->redirect != NULL)
      format = acl->redirect;
   else
   {
      if (defaultAcl == NULL)
	 /* TO-DO: copy the redirect rule for the security or proxy category */
	 /* TO-DO: instead of referring to cgibin.urlfilterdb.com */
	 format = (char *) "http://cgibin.urlfilterdb.com/cgi-bin/URLblocked.cgi?"
		           "clientgroup=%s&category=%t&url=%u";
      else
      {
	 format = defaultAcl->redirect;
	 if (format == NULL)
	    format = (char *) "http://cgibin.urlfilterdb.com/cgi-bin/URLblocked.cgi?"
			      "clientgroup=%s&category=%t&url=%u";
      }
   }

   sgParseRedirect( format, si, acl, si->aclpass, redirect, strcmp(si->protocol,"https")==0 );
}


UFDB_GCC_INLINE
static int hasNoneTerminator( struct AclCategory * ac )
{
   while (ac != NULL)
   {
      if (ac->type == ACL_TYPE_TERMINATOR  &&  !ac->access)
         return 1;
      ac = ac->next;
   }
   return 0;
}


UFDB_GCC_INLINE
static int CategoryInAcl( struct Acl * acl, struct Category * category )
{
   struct AclCategory * ac;
   struct Category * cat;

   for (ac = acl->pass;  ac != NULL;  ac = ac->next)
   {
      cat = ac->cat;
      if (cat != NULL  &&  cat->active  &&  strcmp( cat->name, category->name ) == 0)
         return 1;
   }

   return 0;
}


UFDB_GCC_INLINE
static void addCatToImplicit( struct Acl * acl, struct Category * cat )
{
   struct AclCategory *  ac;
   struct AclCategory ** implicit;

   ac = (struct AclCategory *) ufdbMalloc( sizeof(struct AclCategory) );
   ac->name = ufdbStrdup( cat->name );
   ac->cat = cat;
   ac->access = UFDB_ACL_ACCESS_ALLOW;
   ac->type = ACL_TYPE_DEFAULT;
   ac->nblocks = 0;
   ac->nmatches = 0;
   ac->next = NULL;

   implicit = &(acl->implicitPass);
   while (*implicit != NULL)
   {
      implicit = &((*implicit)->next);
   }
   *implicit = ac;
}


void BuildImplicitPassLists( void )
{
   struct Acl *      acl;
   struct Category * cat;

   /*
    * Build the implicitPass list for all ACLs.
    */
   for (acl = AclList;  acl != NULL;  acl = acl->next)
   {
      acl->hasTerminatorNone = hasNoneTerminator( acl->pass );
      if (!acl->hasTerminatorNone)
      {
	 for (cat = Cat;  cat != NULL;  cat = cat->next)
	 {
	    if (cat->domainlist != NULL  &&  strcmp( cat->name, "checked" ) != 0  &&  !CategoryInAcl( acl, cat ))
	       addCatToImplicit( acl, cat );
	 }
      }
   }
}


UFDB_GCC_HOT 
int UFDBdecideBasedOnSpecificCategory( 
   struct Acl *         acl,
   struct AclCategory * aclpass, 
   struct SquidInfo *   si,
   UFDBrevURL *         revURL,
   char *               redirectURL,
   int                  matchRE  )
{
   int                  decision;
   struct Category *    category;

   decision = UFDB_ACL_ACCESS_DUNNO;
   category = aclpass->cat;
   if (category == NULL  ||  !category->active)
      return decision;

   /* domain name lookup for this source */
   if (category->domainlistDb != NULL) 
   {
      int result;
      int use_rw_lock;
      struct UFDBmemTable * mt;

      use_rw_lock = (category->execdomainlist != NULL);
      if (use_rw_lock)					/* LOCK ====================================== */
      {
	 int ret;
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	 ret = pthread_rwlock_rdlock( &TheDynamicCategoriesLock );
	 if (ret != 0)
	    ufdbLogError( "UFDBdecideBasedOnSpecificCategory: pthread_rwlock_rdlock TheDynamicCategoriesLock failed with code %d", ret );
#else
	 ret = pthread_mutex_lock( &TheDynamicCategoriesLock );
	 if (ret != 0)
	    ufdbLogError( "UFDBdecideBasedOnSpecificCategory: pthread_mutex_lock TheDynamicCategoriesLock failed with code %d", ret );
#endif
      }

      mt = (struct UFDBmemTable *) category->domainlistDb->dbcp;
      result = UFDBlookupRevUrl( &(mt->table.nextLevel[0]), revURL );

      if (use_rw_lock)					/* UNLOCK ==================================== */
      {
	 int ret;
#ifdef UFDB_HAVE_NATIVE_RWLOCK_MUTEX
	 ret = pthread_rwlock_unlock( &TheDynamicCategoriesLock );
	 if (ret != 0)
	    ufdbLogError( "UFDBdecideBasedOnSpecificCategory: pthread_rwlock_unlock TheDynamicCategoriesLock failed with code %d", ret );
#else
	 ret = pthread_mutex_unlock( &TheDynamicCategoriesLock );
	 if (ret != 0)
	    ufdbLogError( "UFDBdecideBasedOnSpecificCategory: pthread_mutex_unlock TheDynamicCategoriesLock failed with code %d", ret );
#endif
      }

      if (result)
      {
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "   UFDBdecideBasedOnSpecificCategory: acl \"%s\" with category \"%s\": returned a match by domain",
			    acl->name, category->name  );

	 si->blockReason = UFDB_API_BLOCKR_ACL;
	 si->aclpass = aclpass;
	 decision = aclpass->access;
	 if (decision == UFDB_ACL_ACCESS_BLOCK)
	    UFDBmakeRedirectString( si, acl, category, redirectURL );
	 return decision;
      }
   }

   /* regular expression lookup for this category */
   if (decision == UFDB_ACL_ACCESS_DUNNO  &&  matchRE  &&  category->regExp != NULL)
   {
      int result;

      result = ufdbRegExpMatch( category->regExp, si->surl );
      if (result) 
      {
	 if (UFDBglobalDebug > 1  ||  UFDBglobalDebugRegexp)
	    ufdbLogMessage( "   UFDBdecideBasedOnSpecificCategory:  acl \"%s\" with category \"%s\": returned a match by expression",
			    acl->name, category->name  );
	 si->blockReason = UFDB_API_BLOCKR_ACL;
	 si->aclpass = aclpass;
	 decision = aclpass->access;
	 return decision;
      }
   }

   /* category SafeSearch */
   if (decision == UFDB_ACL_ACCESS_DUNNO  &&  category->options & UFDB_OPT_SAFE_SEARCH)
   {
      if (UFDBaddSafeSearch( si->domain, si->surl, si->orig )
	      == UFDB_API_MODIFIED_FOR_SAFESEARCH)
      {
	 si->blockReason = UFDB_API_BLOCKR_SAFESEARCH;
	 si->aclpass = aclpass;
	 strcpy( redirectURL, si->orig );
#if UFDB_DEBUG_ACL_ACCESS
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "   UFDBCheckACLaccess  category option SafeSearch matched, return ALLOW" );
#endif
	 decision = UFDB_ACL_ACCESS_ALLOW;
	 return decision;
      }
   }

   return decision;
}


int UFDBdecideAccessBasedOnCategory( 
   struct Acl * acl, 
   struct SquidInfo * si, 
   char * category, 
   char * redirectURL )
{
   struct AclCategory * aclpass;

   for (aclpass = acl->pass;  aclpass != NULL;  aclpass = aclpass->next)
   {
      struct Category * cat;

      cat = aclpass->cat;
      if (cat != NULL  &&  cat->active  &&  strcmp( cat->name, category ) == 0)
      {
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "UFDBdecideAccessBasedOnCategory: category %s %sallowed", 
	                    category, 
			    aclpass->access == UFDB_ACL_ACCESS_BLOCK ? "not " : "" );
	 if (!aclpass->access)
	    UFDBmakeRedirectString( si, acl, cat, redirectURL );
	 return aclpass->access == 0 ? UFDB_ACL_ACCESS_BLOCK : UFDB_ACL_ACCESS_ALLOW;
      }
   }

   for (aclpass = acl->implicitPass;  aclpass != NULL;  aclpass = aclpass->next)
   {
      struct Category * cat;

      cat = aclpass->cat;
      if (cat != NULL  &&  cat->active  &&  strcmp( cat->name, category ) == 0)
      {
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "UFDBdecideAccessBasedOnCategory: category %s (implicitly) allowed", category );
         return UFDB_ACL_ACCESS_ALLOW;
      }
   }

   return UFDB_ACL_ACCESS_DUNNO;
}


UFDB_GCC_HOT
int UFDBurlIsCheckedCategory( struct SquidInfo * si )
{
   if (UFDBglobalCheckedDB.table.nNextLevels > 0)
   {
      if (UFDBlookupRevUrl( &(UFDBglobalCheckedDB.table.nextLevel[0]), si->revUrl ))
      {
	 si->blockReason = UFDB_API_BLOCKR_CHECKED;
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "   UFDBurlIsCheckedCategory: implicit category \"checked\" was matched for URL %s", 
	                    si->url2display );
	 return 1;
      }
      if (UFDBglobalCheckedExpressions != NULL)
      {
	 if (ufdbRegExpMatch( UFDBglobalCheckedExpressions, si->url2display ) == UFDB_API_MATCH)
         {
	    si->blockReason = UFDB_API_BLOCKR_CHECKED;
	    if (UFDBglobalDebug > 1)
	       ufdbLogMessage( "   UFDBurlIsCheckedCategory: implicit category \"checked\" was matched for URL %s", 
			       si->url2display );
	    return 1;
	 }
      }
   }
   return 0;
}


/*
 * decide-implicit-based-on-url:
 *      Decide on dunno/allow based on the URL and the implicitpass list of the ACL.
 *      The decision is based on the URL tables and not expressions.
 *
 * Return value:
 * UFDB_ACL_ACCESS_DUNNO:  don't know.
 * UFDB_ACL_ACCESS_ALLOW:  explicit allow.
 */
UFDB_GCC_HOT
int UFDBdecideImplicitAccessBasedOnURL(
   int                  reconfiguring,
   struct Acl *         acl, 
   struct SquidInfo *   si,
   char *               redirectURL )
{
   int                  decision;
   struct Category *    category;
   struct AclCategory * aclpass;

   if (reconfiguring)
      return UFDB_ACL_ACCESS_DUNNO;

#if 0
   if (UFDBglobalDebug > 1)
      ufdbLogMessage( "   UFDBdecideImplicitAccessBasedOnURL :" );
#endif

   for (aclpass = acl->implicitPass;  aclpass != NULL;  aclpass = aclpass->next)	/* walk all other not-named categories */
   {
      category = aclpass->cat;
      if (category != NULL  &&  !category->active)
	 continue;

      if (UFDBglobalDebug > 1  &&  category != NULL)
	 ufdbLogMessage( "   UFDBdecideImplicitAccessBasedOnURL: testing acl \"%s\" with implicit category \"%s\" for URL %s", 
			 acl->name, category->name, si->url2display );

      decision = UFDBdecideBasedOnSpecificCategory( acl, aclpass, si, si->revUrl, redirectURL, 0 );
      if (decision != UFDB_ACL_ACCESS_DUNNO)
      {
	 si->blockReason = UFDB_API_BLOCKR_PASS;			/* implicit ACL is always "PASS" */
	 si->aclpass = aclpass;
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "   UFDBdecideImplicitAccessBasedOnURL: acl \"%s\" matched implicit category \"%s\" for URL %s", 
	                    acl->name, aclpass->name, si->url2display );
	 return UFDB_ACL_ACCESS_ALLOW;
      }
   }

   if (UFDBurlIsCheckedCategory( si ))
   {
      if (UFDBglobalDebug > 1)
	 ufdbLogMessage( "   UFDBdecideImplicitAccessBasedOnURL: implicit category \"checked\" was matched in acl %s for URL %s", 
			 acl->name, si->url2display );
      return UFDB_ACL_ACCESS_ALLOW;
   }

   return UFDB_ACL_ACCESS_DUNNO;
}


/*
 * decide-based-on-url:
 *      Decide on block/allow based on the URL and the ACL.
 *      The decision is based on the URL tables and expressions, so there is no HTTPS peeking here.
 *	This function also includes checking for the safesearch-category.
 *
 * Return value:
 * UFDB_ACL_ACCESS_DUNNO:  don't know.
 * UFDB_ACL_ACCESS_ALLOW:  explicit allow.
 * UFDB_ACL_ACCESS_BLOCK:  explicit block; a redirection string is returned also.
 */
UFDB_GCC_HOT
int UFDBdecideAccessBasedOnURL(
   int                  reconfiguring,
   struct Acl *         acl, 
   struct SquidInfo *   si, 
   char *               redirectURL )
{
   int                  decision;
   char *               redirect;
   struct sgRewrite *   rewrite = NULL;
   struct AclCategory * aclpass = NULL;
   struct Category *    category = NULL;
   char                 newstring[UFDB_MAX_URL_LENGTH];
   char                 newredir[UFDB_MAX_URL_LENGTH + UFDB_MAX_URL_LENGTH];

   si->blockReason = UFDB_API_BLOCKR_PASS;
   si->aclpass = NULL;
   si->matchedAny = 0;
   si->matchedNone = 0;

#if 0
   if (UFDBglobalDebug > 1)
      ufdbLogMessage( "   UFDBdecideAccessBasedOnURL :" );
#endif

   if (reconfiguring)
   {
#ifdef _POSIX_PRIORITY_SCHEDULING
      static volatile int yieldcounter = 0;
      if (++yieldcounter > 8)
      {
         yieldcounter = 0;
         sched_yield();		
      }
#endif
      si->blockReason = UFDB_API_BLOCKR_LOADING_DB;
      if (UFDBglobalURLlookupResultDBreload == UFDB_DENY)
      {
         strcpy( redirectURL, UFDBglobalLoadingDatabaseRedirect );
         strcpy( si->matchedBy, "loading-db" );
         return UFDB_ACL_ACCESS_BLOCK;
      }
      *redirectURL = '\0';
      return UFDB_ACL_ACCESS_ALLOW;
   }

   if (UFDBglobalFatalError)
   {
      si->blockReason = UFDB_API_BLOCKR_FATAL_ERROR;
      if (UFDBglobalURLlookupResultFatalError == UFDB_DENY)
      {
         strcpy( redirectURL, UFDBglobalFatalErrorRedirect );
         strcpy( si->matchedBy, "fatal-error" );
         return UFDB_ACL_ACCESS_BLOCK;
      }
      *redirectURL = '\0';
      return UFDB_ACL_ACCESS_ALLOW;
   }

   if (acl == NULL)
   {
#if UFDB_DEBUG_ACL_ACCESS
      if (UFDBglobalDebug > 1)
         ufdbLogMessage( "   UFDBdecideAccessBasedOnURL  acl = NULL,  return DUNNO" );
#endif
      return UFDB_ACL_ACCESS_DUNNO;
   }

   si->aclpass = NULL;
   if (acl->pass == NULL)
      acl->pass = defaultAcl->pass;

   if (acl->pass == NULL)
   {  
      decision = UFDB_ACL_ACCESS_BLOCK;
      if (defaultAcl == NULL  ||  defaultAcl->redirect == NULL)
         redirect = (char *) "http://cgibin.urlfilterdb.com/cgi-bin/URLblocked.cgi?"
		             "remark=nodefaultacl&category=%t&url=%u";
      else
	 redirect = defaultAcl->redirect;
   }
   else
   {
      decision = UFDB_ACL_ACCESS_DUNNO;
      redirect = NULL;

      for (aclpass = acl->pass;  aclpass != NULL;  aclpass = aclpass->next)	/* walk all explicitly named categories */
      {
         category = aclpass->cat;
         if (category != NULL  &&  !category->active)
  	    continue;

#if UFDB_DEBUG_ACL_ACCESS
         if (UFDBglobalDebug > 2  &&  category != NULL)
  	    ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: testing acl \"%s\" with category \"%s\" for %s %s",
	                    acl->name, category->name, si->method, si->url2display );
#endif

         if (aclpass->type == ACL_TYPE_TERMINATOR)  				/*  pass any/none  */
         {
            si->aclpass = aclpass;
	    if (aclpass->access)	/* ANY */
	    {
	       /* Be careful here: "any" is matched, but if port==443 or method==CONNECT we must do HTTPS probing as well 
	        * to decide if the URL is allowed or not.
		*/
	       if (si->port == 80)
	       {
		  if (UFDBglobalDebug > 1)
		     ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: pass; acl \"%s\" matched pseudo category \"any\" port=80 (PASS) for %s %s",
		                     acl->name, si->method, si->url2display );
		  si->matchedAny = 1;
		  si->blockReason = UFDB_API_BLOCKR_PASS;
	       }
	       else
	       {
		  /* here we cannot make a decision about block/pass since we might need a HTTPS probe for that decision */
		  /* TODO: why??? isn't the URL PASSed ? */
		  if (UFDBglobalDebug > 1)
		     ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: dunno; acl \"%s\" matched pseudo category \"any\" port!=80 (UNDECIDED) for %s %s",
		                     acl->name, si->method, si->url2display );
		  si->matchedAny = 2;
	       }
	    }
	    else			/* NONE */
	    {
	       if (UFDBglobalDebug > 1)
	          ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: block; acl \"%s\" matched pseudo category \"none\" (BLOCK) for %s %s",
		                  acl->name, si->method, si->url2display );
	       decision = UFDB_ACL_ACCESS_BLOCK;
	       si->matchedNone = 1;
	       si->blockReason = UFDB_API_BLOCKR_ACL_NONE;
	    }
	    break;
         }
	 else
         if (aclpass->type == ACL_TYPE_INADDR) 					/*  pass [!]in-addr  */
         {
	    if (si->isip) 
	    {
	       si->aclpass = aclpass;
	       decision = aclpass->access;
	       if (decision == UFDB_ACL_ACCESS_BLOCK)
		  si->blockReason = UFDB_API_BLOCKR_ACL;
	       if (UFDBglobalDebug > 1)
	          ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: block; acl \"%s\" matched pseudo category \"in-addr\" for %s %s",
		                  acl->name, si->method, si->url2display );
	       break;
	    }
	    continue;
         }
  
	 decision = UFDBdecideBasedOnSpecificCategory( acl, aclpass, si, si->revUrl, redirectURL, 1 );	/*  pass [!]category  */
	 if (decision != UFDB_ACL_ACCESS_DUNNO)
	 {
	    si->aclpass = aclpass;
	    if (decision == UFDB_ACL_ACCESS_BLOCK)
	       si->blockReason = UFDB_API_BLOCKR_ACL;
	    if (UFDBglobalDebug > 1)
	       ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: %s; acl \"%s\" matched category \"%s\" for  port %d  protocol %s  method %s  URL %s", 
	                       decision == UFDB_ACL_ACCESS_BLOCK ? "block" : "pass", 
			       acl->name, aclpass->name, si->port, si->protocol, si->method, si->url2display );
	    break;
	 }
	 /* TO-DO: look at rtsp and mms protocols */
      }

      if (decision == UFDB_ACL_ACCESS_DUNNO)
      {
	 if (si->port == 80)
	 {
	    for (aclpass = acl->implicitPass;  aclpass != NULL;  aclpass = aclpass->next)	/* walk all other not-named categories */
	    {
	       category = aclpass->cat;
	       if (category != NULL  &&  !category->active)
		  continue;
	
#if UFDB_DEBUG_ACL_ACCESS
	       if (UFDBglobalDebug > 2  &&  category != NULL)
		  ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: testing acl \"%s\" with implicit category \"%s\" for %s %s",
		                  acl->name, category->name, si->method, si->url2display );
#endif

	       decision = UFDBdecideBasedOnSpecificCategory( acl, aclpass, si, si->revUrl, redirectURL, 0 );
	       if (decision != UFDB_ACL_ACCESS_DUNNO)
	       {
		  si->blockReason = UFDB_API_BLOCKR_PASS;			/* implicit ACL is always "PASS" */
	          si->aclpass = aclpass;
		  if (UFDBglobalDebug > 1)
		     ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: pass; acl \"%s\" matched implicit category \"%s\" for %s %s",
		                     acl->name, aclpass->name, si->method, si->url2display );
		  break;
	       }
	    }
	 }
	 else
	 {
#if 0
	    if (UFDBglobalDebug > 1)
	       ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: acl %s: matching against implicit categories is skipped since port=%d", 
	                       acl->name, si->port );
#endif
	 }
      }

      if (decision == UFDB_ACL_ACCESS_DUNNO)
      {
         if (UFDBurlIsCheckedCategory( si ))
	 {
	    static struct AclCategory static_aclpass_checked = { 0, "checked", NULL, 0, 0, NULL, "", 0 };

	    decision = UFDB_ACL_ACCESS_ALLOW;
	    si->aclpass = &static_aclpass_checked;
	    if (UFDBglobalDebug > 1)
	       ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: pass; acl \"%s\" matched implicit category \"checked\" for %s %s",
			       acl->name, si->method, si->url2display );
	 }
      }

      if (decision == UFDB_ACL_ACCESS_BLOCK) 
      {
         if (category != NULL  &&  category->redirect != NULL)
	    redirect = category->redirect;
         else if (category != NULL  &&  category->rewrite != NULL  &&  strcmp( si->protocol, "https" ) != 0  &&
  	          (redirect = sgRewriteExpression( category->rewrite, si->orig, newstring )) != NULL)
         {
  	    ;
         }
         else if (acl->redirect == NULL)
	    redirect = defaultAcl->redirect;
         else
	    redirect = acl->redirect;
         if (redirect == NULL)
         {
#if UFDB_DEBUG_ACL_ACCESS
            /* suppress many errors. we only want it once! */
            ufdbLogError( "!!!!! there is no \"redirect\" rule for ACL %s/%s  *****", 
  		          acl->name, aclpass->name );
#endif
            /* TO-DO: copy the redirect rule for the security or proxy category */
            /* TO-DO: instead of referring to cgibin.urlfilterdb.com */
            redirect = (char *) "http://cgibin.urlfilterdb.com/cgi-bin/URLblocked.cgi?"
  	                        "clientgroup=%s&category=%t&url=%u";
         }
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: block; redirect is now '%s'", redirect );
      }
   }

   if (decision == UFDB_ACL_ACCESS_BLOCK)
   {
      if (UFDBglobalSquidUsesSSLbump  &&  strcmp( si->method, "CONNECT" ) == 0)
      {
	 /* For some categories it is necessary to block the CONNECT since the server does not have SSL+HTTP */
	 if (category != NULL  &&  category->blockBumpedConnect)
	 {
	    si->BlockedBumpedConnectAllowed = 0;
	    redirect = UFDBglobalRedirectBumpedHttps;
	    if (UFDBglobalDebug > 1)
	       ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: block; %s %s redirected to %s since block-bumped-connect is on", 
	                       si->method, si->url2display, redirect );
	 }

	 /* squid-ssl-bump + CONNECT : 
	  * the default decision is to allow the CONNECT and the block the next GET/POST, *BUT*
	  * this does not work for CONNECT to non-HTTP-based sites where the CONNECT itself must be blocked.
	  * So probe the server to see if it uses HTTP which we subsequently can block.
	  * Note that a probe is not necessary for Teamviewer, Skype and others.  TODO: skip probe for these sites.
	  */
	 /* TODO: if (si->sni[0] != '\0' || (serverProto = probeServer[Cached](si->domain)) != HTTP) { block-this-URL } */
	 if (UFDBglobalDebug > 1)
	    ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: dunno; blocked CONNECT to '%s' is skipped because squid-uses-active-bumping; SNI '%s'", si->orig, si->sni );
	 redirect = NULL;
         decision = UFDB_ACL_ACCESS_DUNNO;	/* do not return ALLOW since we might want to check the certificate */
	 si->BlockedBumpedConnectAllowed = 1;
      }
      else
      {
	 /* TO-DO: what if CONNECT to port other then 443 ? */
	 if (si->port == 443  ||  strcmp( si->protocol, "https" ) == 0)
	 {
	    if (strcmp( si->method, "CONNECT" ) == 0  ||  !UFDBglobalSquidUsesSSLbump)
	       redirect = UFDBglobalRedirectHttps;
	    else
	       redirect = UFDBglobalRedirectBumpedHttps;
	    if (UFDBglobalDebug > 1)
	       ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: block; %s %s redirected to %s", si->method, si->url2display, redirect );
	 }
      }
   }

   if (acl->rewrite == NULL)
      rewrite = defaultAcl->rewrite;
   else
      rewrite = acl->rewrite;
   if (rewrite != NULL  &&  decision == UFDB_ACL_ACCESS_DUNNO)
   {
      redirect = sgRewriteExpression( rewrite, si->orig, newstring );
   }
   else if (redirect != NULL) 
   {
      redirect = sgParseRedirect( redirect, si, acl, aclpass, newredir, strcmp(si->protocol,"https")==0 );
   }

   if (redirect != NULL)
   {
      strcpy( redirectURL, redirect );
      si->aclpass = aclpass;
      if (UFDBglobalDebug > 1)
         ufdbLogMessage( "   UFDBdecideAccessBasedOnURL: block;  return BLOCK  redirect: %s", redirectURL );
      return UFDB_ACL_ACCESS_BLOCK;
   }

   *redirectURL = '\0';
#if UFDB_DEBUG_ACL_ACCESS
   if (UFDBglobalDebug > 1)
      ufdbLogMessage( "   UFDBdecideAccessBasedOnURL  end, return %s", decision == UFDB_ACL_ACCESS_DUNNO ? "DUNNO" : "PASS" );
#endif

   return decision;
}


UFDB_GCC_HOT
int UFDBdecideAccessBasedOnHostname(
   int                   reconfiguring,
   struct Acl *          acl, 
   struct SquidInfo *    si, 
   UFDBthreadAdmin *     admin,
   char *                hostname,
   char *                redirectURL )
{
   struct AclCategory *  ac;
   UFDBrevURL *          revurl;
   int                   decision;

   if (reconfiguring)
      return UFDB_ACL_ACCESS_DUNNO;

   if (*hostname == '*'  &&  *(hostname+1) == '.')
      hostname += 2;
   else if (*hostname == 'w'  &&  *(hostname+1) == 'w'  &&  *(hostname+2) == 'w'  &&  *(hostname+3) == '.')
      hostname += 4;

   if (UFDBglobalDebug > 1)
      ufdbLogMessage( "   UFDBdecideAccessBasedOnHostname: decision is based on hostname \"%s\"", hostname );

   if (*hostname == '\0')
      return UFDB_ACL_ACCESS_DUNNO;

   decision = UFDB_ACL_ACCESS_DUNNO;
   revurl = UFDBgenRevURL( admin, (unsigned char *) hostname );		/* note: cannot use si->revurl */

   for (ac = acl->pass;  ac != NULL;  ac = ac->next)
   {
      decision = UFDBdecideBasedOnSpecificCategory( acl, ac, si, revurl, redirectURL, 0 );
      if (decision != UFDB_ACL_ACCESS_DUNNO)
      {
	 si->aclpass = ac;
	 break;
      }
   }

   if (decision == UFDB_ACL_ACCESS_DUNNO)
   {
      for (ac = acl->implicitPass;  ac != NULL;  ac = ac->next)
      {
	 decision = UFDBdecideBasedOnSpecificCategory( acl, ac, si, revurl, redirectURL, 0 );
	 if (decision != UFDB_ACL_ACCESS_DUNNO)
	 {
	    si->aclpass = ac;
	    break;
	 }
      }
   }

   if (UFDBglobalDebug > 1)
   {
      if (ac == NULL)
	 ufdbLogMessage( "   UFDBdecideAccessBasedOnHostname: hostname %s did not match any category", hostname );
      else
	 ufdbLogMessage( "   UFDBdecideAccessBasedOnHostname: hostname %s matched by category %s, decision is %s",
	                 hostname, ac->name, decision == UFDB_ACL_ACCESS_BLOCK ? "block" : "allow" );
   }

   UFDBfreeRevURL( admin, revurl );

   return decision;
}

